package Program;

use strict;
use vars qw/$VERSION @ISA @EXPORT/;

require Exporter;
require AutoLoader;
require Program::Statement;

@ISA = qw(Exporter AutoLoader);
# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.
@EXPORT = ();
$VERSION = '0.01';

# Preloaded methods go here.

sub new {
    my $that = shift;
    my $class = ref($that) || $that;
    my $progname = shift;
    my $self = {
	Program_name => $progname,
	First_statement => undef,
	Current_statement => undef,
	Language => undef,
	Text => undef
    };
    bless $self, $class;

    # Determine what language the program is.
    $self->{Language} = $self->determine_language;
    if ($self->{Language}) {
	my $str = $self->{Language};
	$str =~ s/\+/plus/g;
	eval "require Program::$str; \$self = Program::$str->new(\$progname)";
    }

    return $self;

}

sub initialize_and_parse {
    my $self = shift;
    $self->initialize;
    $self->parse;
}

# Read the program text from the program's file.
sub initialize {
    my $self = shift;
    unless ($self->{Program_name}) {
	$self->{Program_name} = shift;
	return 0 unless defined $self->{Program_name};
    }
    open PROGRAM, $self->{Program_name};
    {
	local $/ = undef;
	$self->{Text} = <PROGRAM>;
    }
    close PROGRAM;

}

sub parse {
    my $self = shift;
    die "Language not defined.\n" unless defined $self->{Language};

    $self->{First_statement} = Program::Statement->new();
    $self->{Current_statement} = $self->{First_statement};
    local $_ = $self->{Text};
    my @brackets = ();

    my @regexps = qw/./;
    my $type;
    for $type (qw/Other PreProcessor Comment Quote/){
	unshift @regexps, $self->{$type} if defined $self->{$type};
    }
    my $regexp = '(' . (join '|', @regexps) . ')';

    while (/$regexp/smog) { 
	my $statement = $self->{Current_statement};
	my $found = $1;

	if (defined $self->{'Quote'} and $found =~ /^($self->{'Quote'})$/sm) {
            $statement->add_quoted_code($found);
	}
	elsif (defined $self->{'Comment'} and
	       $found =~ /^(?:$self->{'Comment'})$/sm) {
	    my $comment;
	    {
		no strict 'refs';
		my $ctr = 0;
		until(defined ($comment = $ {++$ctr}) or $ctr > 20){}
	    }
	    # Remove leading *'s, #'s , or -'s from the comment.
	    # Are there other frequently occurring characters like this?
	    $comment =~ s/(?:^|\n)\s*[-*#]+/ /mg;
	    $statement->add_comment($comment);
	    $statement->add_code(' ');
        }
	elsif (defined($self->{'PreProcessor'}) and
	       $found =~ /^($self->{'PreProcessor'})$/sm) {
	    # Do nothing for now; May want to either move to next
	    # statement or try to figure out what is going on.
        }
	# I don't think the same sort of thing is required for 'Other' -
	# The code will be added in the "else" clause at the bottom.

	# Is this an open bracket?
        elsif ((index $self->{'Brackets'}, $found) >= 0
	       and length($found) == 1) {
	    push @brackets, $found;
	    $statement->add_code($found);
	}
	elsif (@brackets) {	# If there are remaining open brackets,
				# then the statement isn't complete.
	    $statement->add_code($found);
	    # Future improvements:
	    # 1. Check @brackets later on to see if the brackets match.
	    # 2. Modularize the '(' '[' '{' strings below.
	    pop @brackets
		if $found eq '}' or $found eq ']' or $found eq ')';
	    push @brackets, $found
		if $found eq '{' or $found eq '[' or $found eq '(';
	}
	elsif (((index $self->{'EndOfStmt'}, $found) >= 0
		and length($found) == 1)
	       and $self->is_end_of_stmt) {
	    $self->{Current_statement} = $statement->next_statement;
	}
	elsif (((index $self->{'BlockBrackets'}, $found) >= 0
	       and length($found) == 1)) {
	    if ($self->code_starts_block) {
		$self->{Current_statement} = $statement->up;
	    }
	    else {
		push @brackets, $found;
		$statement->add_code($found);
	    }
	}
	elsif ((index &Program::matching_brackets($self->{'BlockBrackets'}),
		$found) >= 0) {
	    if (ref $statement->get_down) {
		$self->{Current_statement} = $statement->last_statement;
	    }
	    else {		# Mismatched brackets; soldier on...
		$self->{Current_statement} = $statement->next_statement;
	    }
	}
	else { $statement->add_code($found) }
   	$statement->add_text($found);
    }
}

sub is_end_of_stmt{1}
sub code_starts_block{1}

# Returns the bracket that would match the bracket passed as a parameter.
# Note that this is *not* a method!  Ergo, it might fit better elsewhere.
sub matching_brackets {
    my $string = shift;
    $string =~ y/()[]{}/)(][}{/;
    return $string;
}

# Create an inheritable print_sub_descriptions, later.  This should only
# be a matter of creating a stmt_is_subr method in each subclass that
# returns the name of a subroutine if this current statement is the start
# of a subroutine, and undef if it isn't.  Sounds simple, eh?

# Determine the program language.
sub determine_language {
    my $self = shift;
    my ($extension) = $self->{Program_name} =~ /.*\.(.*)/;
    my %xmap = (C => '[ch]',
		TCL => 'tcl',
		Perl => '(?:p[hlm]|PL|al)',
		Java => 'java?',
		'C++' => '(?:cc|hh|cpp)',
		Lisp => '(?:[ce]l|lisp)');
    for (keys %xmap) {return $_ if $extension =~ /^$xmap{$_}$/}
    return undef;
}

# Print descriptions of subroutines.
sub print_sub_descriptions {
    my $self = shift;
    my $func = shift;		# After this, @_ holds all remaining args.
    my $statement = $self->{First_statement};
    my $subname;
    
    while (ref $statement) {
	&$func($subname, $statement->{code}, $statement->{description},
	       $statement->{comment}, $self->{Program_name},
	       $self->{Language}, @_)
	    if ($subname = $self->stmt_is_subr($statement));
        $statement = $statement->next_statement_in_code;
    }
}

# This subroutine is mostly for testing.
sub traverse_all_statements {
    my $self = shift;
    my $statement = $self->{First_statement};
    while (ref $statement) {
#	if (!(ref $statement->get_down) and $statement->get_code =~ /\) ?/) {
	    print $statement->get_code, ':', $statement->get_comment, "\n";
#	}
	$statement = $statement->next_statement_in_code;
    }
}

# Autoload methods go after =cut, and are processed by the autosplit program.

1;
__END__

# Below is the stub of documentation for your module. You better edit it!

=head1 NAME

Program - An object representation of a computer program

=head1 DESCRIPTION

The Program object is a package of routines that deal with source
code for computer programs.

Program provides an object for working with program source code
files, particularly parsing that code and extracting comments,
code, etc.  Program's methods (except for new) are generally not
called directly, but rather are inherited by Program::I<language>,
where I<language> is an object for dealing with program code
written in I<language>.

Program has the following methods:

=over 2

=item new

Called with the parameter of the name of the program, returns an
object representing that program.  The object returned will be
of class Program::I<language>, where I<language> is the language
that the new method believed the program was.

=item initialize

Reads the program text from the specified file.

=item parse

Parses the text, assuming you have already called initialize().
The default parse routine just parses the text into literals,
comments, code, etc. so if you want something more detailed
you probably don't want to inherit this method.

=item initialize_and_parse

Same as calling both the initialize() and parse() methods.

=item determine_language

Determines the language of the program.

=item print_sub_descriptions

Prints descriptions for each subroutine in the parsed program.

=item traverse_all_statements

Traverses all statements and prints them out.  Used only for testing.

=back

=head1 BUGS

determine_language currently only checks the program extension,
which doesn't always help.  A system() call to file might help.

The parser isn't (nor is it intended to be) complex enough to
create, for example, compilers from it.

There is currently little support for the Documentation property.

=head1 AUTHOR

James Yolkowski (ajy@stormloader.com).

=head1 SEE ALSO

perl(1).

=cut
